/*  GNU Ocrad - Optical Character Recognition program
    Copyright (C) 2003 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <algorithm>
#include <cstdio>
#include <vector>
#include "common.h"
#include "rectangle.h"
#include "block.h"
#include "blockmap.h"
#include "profile.h"
#include "feats.h"
#include "iso_8859_1.h"


Features::Features( const Block & b ) throw()
  : _block( &b ), _blockmap( b.blockmap() ), _hbars( -1 ), _vbars( -1 ),
    hscan_valid( false ), vscan_valid( false ), lp( b, Profile::left ),
    tp( b, Profile::top ), rp( b, Profile::right ), bp( b, Profile::bottom ),
    hp( b, Profile::height ), wp( b, Profile::width )
  {}


int Features::hbars() const throw()
  {
  if( _hbars < 0 )
    {
    const Block & b = *_block;
    const Blockmap & bm = *_blockmap;
    int state = 0, begin = 0, l = 0, r = 0, limit = wp.max() / 2;
    int count[b.height()];
    _hbars = 0;

    for( int row = b.top(); row <= b.bottom(); ++row )
      {
      int col, c = 0, lt = 0, rt = 0, x = 0;
      int & maxcount = count[row-b.top()];
      maxcount = 0;
      for( col = b.left(); col <= b.right(); ++col )
        {
        if( bm.id( row, col ) == b.id() )
          { ++c; x = col; if( col < b.right() ) continue; }
        if( c > maxcount ) { maxcount = c; rt = x; lt = rt - c + 1; }
        c = 0;
        }
      switch( state )
        {
        case 0: if( maxcount > limit )
                  { state = 1; begin = row; l = lt; r = rt; }
                else break;
        case 1: if( maxcount > limit )
                  {
                  if( lt < l ) l = lt;
                  if( rt > r ) r = rt;
                  if( row < b.bottom() ) break;
                  }
                state = 0;
                int end = ( maxcount <= limit ) ? row - 1 : row;
                int width = r - l + 1;
                while( begin <= end && 3 * count[begin-b.top()] < 2 * width )
                  ++begin;
                while( begin <= end && 3 * count[end-b.top()] < 2 * width )
                  --end;
                if( begin > end ) break;
                _hbar.push_back( Rectangle( l, begin, r, end ) );
                ++_hbars; break;
        }
      }
    }
  return _hbars;
  }


int Features::vbars() const throw()	// FIXME small gaps not detected
  {
  if( _vbars < 0 )
    {
    const Block & b = *_block;
    const Blockmap & bm = *_blockmap;
    int state = 0, begin = 0, limit = b.height();
    limit -= ( b.height() < 40 ) ? 3 : b.height() / 10;
    _vbars = 0;

    for( int col = b.left(); col <= b.right(); ++col )
      {
      int c = 0, c2 = 0, count = 0;
      for( int row = b.top() + 1; row < b.bottom(); ++row )
        {
        if( bm.id( row, col ) == b.id() )
          { ++c; if( row < b.bottom() - 1 ) continue; }
        else if( ( col > b.left() && bm.id( row, col - 1 ) == b.id() ) ||
                 ( col < b.right() && bm.id( row, col + 1 ) == b.id() ) )
          { ++c; ++c2; if( row < b.bottom() - 1 ) continue; }
        if( c > count ) { count = c; } c = 0;
        }
      if( ( count - c2 ) * 3 < limit * 2 ) count = 0;
      switch( state )
        {
        case 0: if( count >= limit ) { state = 3; begin = col; }
                else if( count * 4 >= limit * 3 ) { state = 2; begin = col; }
                else if( count * 3 >= limit * 2 ) { state = 1; begin = col; }
                break;
        case 1: if( count >= limit ) state = 3;
                else if( count * 4 >= limit * 3 ) state = 2;
                else if( count * 3 < limit * 2 ) state = 0;
                else begin = col;
                break;
        case 2: if( count >= limit ) state = 3;
                else if( count * 3 < limit * 2 ) state = 0;
                else if( count * 4 < limit * 3 ) state = 1;
                break;
        case 3: if( count * 3 < limit * 2 || col == b.right() )
                  {
                  int end = ( col == b.right() ) ? col : col - 1;
                  _vbar.push_back( Rectangle( begin, b.top(), end, b.bottom() ) );
                  ++_vbars; state = 0;
                  }
        }
      }
    }
  return _vbars;
  }


// return the number of horizontal traces crossing every column
const std::vector< int > & Features::hscan() const throw()
  {
  if( !hscan_valid )
    {
    const Block & b = *_block;
    const Blockmap & bm = *_blockmap;
    int state = 0;
    hscan_valid = true; _hscan.resize( b.width() );

    for( int col = b.left(); col <= b.right(); ++col )
      {
      _hscan[col-b.left()] = 0;
      for( int row = b.top(); row <= b.bottom(); ++row )
        {
        bool black = ( bm.id( row, col ) == b.id() );
        switch( state )
          {
          case 0: if( black ) state = 1; break;
          case 1: if( !black || row == b.bottom() )
                    { state = 0; ++_hscan[col-b.left()]; }
          }
        }
      }
    }
  return _hscan;
  }


// return the number of vertical traces crossing every row
const std::vector< int > & Features::vscan() const throw()
  {
  if( !vscan_valid )
    {
    const Block & b = *_block;
    const Blockmap & bm = *_blockmap;
    int state = 0;
    vscan_valid = true; _vscan.resize( b.height() );

    for( int row = b.top(); row <= b.bottom(); ++row )
      {
      _vscan[row-b.top()] = 0;
      for( int col = b.left(); col <= b.right(); ++col )
        {
        bool black = ( bm.id( row, col ) == b.id() );
        switch( state )
          {
          case 0: if( black ) state = 1; break;
          case 1: if( !black || col == b.right() )
                    { state = 0; ++_vscan[row-b.top()]; }
          }
        }
      }
    }
  return _vscan;
  }


// Looks for three black sections in column hcenter()  1, then tests if
// upper and lower gaps are open to the right or to the left
unsigned char Features::test_235Esz() const throw()
  {
  static const int csize = 5;
  static const int ucoff[csize] = { 0, +1, -1, -1, +1 };
  static const int lcoff[csize] = { 0, +1, -1, +1, -1 };
  const Block & b = *_block;
  const Blockmap & bm = *_blockmap;
  int lrow1 = 0, urow2 = 0, lrow2 = 0, urow3 = 0;
  int lcol1 = 0, ucol2 = 0, lcol2 = 0, ucol3 = 0, black_section = 0;
  int cid = bm.id( b.vcenter(), b.hcenter() );

  for( int i = 0; i < csize; ++i )
    {
    int ucol = b.hcenter() + ucoff[i];
    int lcol = b.hcenter() + lcoff[i];
    if( ucol != lcol &&
        ( bm.id( b.vcenter(), ucol ) != cid || bm.id( b.vcenter(), lcol ) != cid ) )
      continue;
    bool prev_black = false;
    for( int row = b.top(); row <= b.bottom(); ++row )
      {
      int col = ( row <= b.vcenter() ) ? ucol : lcol;
      bool black = ( bm.id( row, col ) == b.id() );
      if( black )
        { if( !prev_black )
          { if( ++black_section == 2 ) { urow2 = row - 1; ucol2 = col; }
          else if( black_section == 3 ) { urow3 = row - 1; ucol3 = col; } } }
      else if( prev_black )
        { if( black_section == 1 ) { lrow1 = row; lcol1 = col; }
          else if( black_section == 2 ) { lrow2 = row; lcol2 = col; } }
      prev_black = black;
      }
    if( black_section == 3 && lrow1 <= b.vpos(30) && urow3 >= b.vpos(70) )
      break;
    black_section = 0;
    }

  if( black_section != 3 ) return 0;
  if( b.escape_left( lrow2, lcol2 ) )
    {
    if( b.escape_left( urow2, ucol2 ) )
      {
      if( b.escape_bottom( urow3, ucol3 ) && b.escape_top( lrow1, lcol1 ) )
        return '*';
      int m = lp.minima(); if( m < 2 || m > 3 ) return 0;
      if( !bp.ispit() && m != 3 ) return 0;
      if( b.height() > 3 * b.width() ) return 0;
      return '3';
      }
    if( b.escape_right( urow2, ucol2 ) )
      {
      if( b.seek_right( lrow1 + 1, lcol1 ) >= b.right() ) return '5';
      if( urow2 > b.vcenter() + 1 && lrow2 > b.bottom() - (b.height() / 5) &&
          b.seek_right( urow2 - 1, ucol2 ) < b.right() )
        return ISO_8859_1::SCCEDI;
      return 's';
      }
    }
  else if( b.escape_right( lrow2, lcol2 ) )
    {
    if( b.escape_right( urow2, ucol2 ) )
      {
      if( b.escape_bottom( urow3, ucol3 ) && b.escape_top( lrow1, lcol1 ) )
        return '*';
      if( bp.minima( b.height() / 5 ) == 1 )
        {
        if( b.escape_bottom( urow3, ucol3 ) )
          { if( 2 * lp[lp.samples()-2] > rp[rp.samples()-2] ) return '';
          return 'F'; }
        if( b.escape_top( urow2, b.left() ) &&
            !b.escape_top( lrow2, b.left() ) ) return 'f';
        else return 'E';
        }
      }
    if( b.escape_left( urow2, ucol2 ) )
      {
      if( tp.isflat() ) return 'z';
      if( ( tp.isconvex() || ( tp.ispit() && !bp.ispit() ) ) &&
          lp[lp.pos(50)] >= b.width() / 2 ) return '2';
      if( b.height() > 2 * wp.max() && bp.minima( b.height() / 5 ) == 1 )
        { if( b.height() >= 3 * wp.max() ) return 'l'; else return 'I'; }
      return 'z';
      }
    }
  return 0;
  }


// Tests if the lower half of character is open to the left, to the right,
// and/or to the bottom
unsigned char Features::test_49ARegpq() const throw()
  {
  const Block & b = *_block;

  if( bp.minima( b.height() / 10 ) == 2 && bp.isctip() )
    { if( tp.isvpit() ) return 'A'; else return 'R'; }

  const Block & h = b.block_vector().front();		// hole
  int col = h.hcenter();
  int row = b.seek_bottom( h.bottom(), col, false ) + 1;
  if( row >= b.vpos(90) )
    { col = h.left(); row = b.seek_bottom( h.bottom(), col, false ) + 1; }
  if( row >= b.bottom() ) return 0;

  if( b.escape_right( row, col ) )
    {
    if( ( lp.ispit() && b.seek_bottom( row, h.right() ) < b.bottom() ) ||
        ( lp.isconvex() && b.seek_bottom( row, h.hcenter() ) < b.bottom() ) )
      return 'e';
    if( bp.ispit() )
      {
      int row2 = b.seek_bottom( row, h.right() );
      if( row2 < b.vpos(75) ) return 'g';
      if( row2 < b.bottom() ) return 'e';
      }
    return 'p';
    }

  if( b.escape_left( row, col ) )
    {
    Profile hlp( h, Profile::left );
    Profile htp( h, Profile::top );
    if( vbars() == 1 && vbar(0).hcenter() > b.hcenter() &&
        hlp.decreasing() && htp.decreasing() )
      return '4';
    if( rp.isconvex() && tp.isconvex() && rp.minima() == 1 )
      return '9';
    int hdiff;
    if( b.bottom_hook( &hdiff ) && hdiff > 0 )
      { if( rp.isctip() ) return 's'; return 'g'; }
    if( row > b.vpos(80) ) return 'Q';
    if( b.seek_bottom( row, col ) < b.bottom() &&
        rp.increasing( row - b.top() ) ) return 'g';
    return 'q';
    }
  return 0;
  }


unsigned char Features::test_4ADQao() const throw()
  {
  const Block & b = *_block;
  const Block & h = b.block_vector().front();

  if( lp.isflats() && !rp.isflats() ) return 'D';
  if( !rp.isconvex() )
    {
    if( tp.minima() == 2 ) return '#';
    if( tp.minima() == 1 && bp.minima() == 1 )
      {
      int row = b.seek_bottom( h.bottom(), h.hcenter(), false );
      if( !lp.isconvex() && bp.isconvex() &&
          b.seek_bottom( row, h.hcenter() ) < b.bottom() )
        return ISO_8859_1::SEACUTE;
      row = ( row + b.seek_bottom( row, h.hcenter() ) ) / 2;
      if( row < b.bottom() - 1 && !lp.isflats() &&
          b.seek_left( row, h.hcenter() ) <= b.left() ) return '4';
      }
    if( b.width() > 3 * h.width() )
      {
      int c = vscan()[b.vcenter() - b.top()];
      if( c == 3 && h.top() < b.vcenter() && !lp.isconvex() ) return 'm';
      if( c >= 4 ) return '@';
      }
    if( tp.minima() == 1 && bp.minima() == 2 && !lp.isconvex() ) return 'A';
    }
  if( Ocrad::similar( h.left() - b.left(), b.right() - h.right(), 40 ) )
    {
    if( !lp.isconvex() && !bp.isconvex() && !rp.isflat() && bp[1] < 2 )
      return 'D';
    if( bp.minima() > 1 || rp.minima() > 1 ) return 'Q';
    return 'o';
    }
  if( h.left() - b.left() > b.right() - h.right() && rp.ispit() ) return 'D';
  if( bp.minima() > 1 || rp.minima() > 1 ) return 'a';
  return 0;
  }


// Tests if the upper half of character is open to the left, to the right,
// and/or to the bottom
unsigned char Features::test_6abd() const throw()
  {
  const Block & b = *_block;
  const Block & h = b.block_vector().front();		// hole

  int col = h.hcenter();
  int row = b.seek_top( h.top(), col, false ) - 1;
  if( row <= b.top() )
    {
    col = h.right(); if( b.right() - h.right() > h.width() ) ++col;
    row = b.seek_top( h.top(), col, false ) - 1;
    }
  if( row <= b.top() ) return 0;
  int rcol = ( b.right() + h.right() ) / 2;
  int urow = h.top() - ( b.bottom() - h.bottom() );
  bool oacute1 = ( ( b.seek_right( urow - 1, h.right() ) >= b.right() ) ||
                   ( b.seek_right( row, col ) >= b.right() ) );

  if( b.escape_right( row, col ) )
    {
    int c = lp[urow-b.top()];
    bool oacute2 = ( c > lp[h.top()-b.top()] + 1 &&
                     tp[std::min(c-1,b.width()/4)] + b.top() >= urow );
    if( oacute1 && oacute2 )
      {
      bool oacute3 = ( b.seek_left( b.top() + 1, b.right() + 1 ) >= rcol ||
                       b.seek_right( h.top(), b.left() - 1 ) <= b.left() + 1 );
      if( oacute3 ) return ISO_8859_1::SOACUTE;
      }
    if( !oacute2 && lp.ispit() && bp.ispit() )
      {
      int row2 = b.seek_top( h.top(), h.right() + 1, false ) - 1;
      row2 = b.seek_top( row2, h.right() + 1 );
      if( row2 > b.top() ) return '6';
      }
    int row2 = b.seek_top( h.top(), rcol, false ) - 1;
    row2 = b.seek_top( row2, rcol );
    if( row2 <= b.top() ) return 'b'; else return 's';
    }

  if( b.escape_left( row, col ) )
    {
    int row2 = b.seek_top( h.top(), h.left(), false ) - 1;
    row2 = b.seek_top( row2, h.left() );
    if( row2 > b.top() )
      { if( b.vcenter() < row ) return ISO_8859_1::SAACUTE; else return 'a'; }
    if( oacute1 ) return ISO_8859_1::SOACUTE; else return 'd';
    }

  if( b.width() > 3 * h.width() && h.top() < b.vcenter() &&
      vscan()[b.vcenter() - b.top()] == 3 && !lp.isconvex() ) return 'm';
  int hdiff; if( b.top_hook( &hdiff ) && hdiff > 0 ) return 's';
  return 0;
  }


unsigned char Features::test_CEFIJLlT() const throw()
  {
  const Block & b = *_block;

  if( tp.minima( b.height() / 4 ) != 1 || bp.minima( b.height() / 4 ) != 1 )
    return 0;

  int row, col;
  if( b.seek_right( b.vcenter(), b.left() - 1 ) >= b.hcenter() )
    col = b.hpos( 25 );
  else col = b.hpos( 75 );
  row = b.seek_top( b.vcenter(), col );
  if( row <= b.top() || ( row < b.vpos( 25 ) && b.escape_top( row, col ) ) )
    {
    int hdiff;
    if( b.bottom_hook( &hdiff ) )
      {
      if( hdiff > b.height() / 2 && rp.increasing( rp.pos(75) ) ) return 'J';
      if( -hdiff > b.height() / 2 ) return 'L';
      }
    }

  if( vbars() == 1 && vbar(0).width() >= 2 &&
      3 * vbar(0).width() <= b.width() + 2 )
    {
    if( std::abs( vbar(0).hcenter() - b.hcenter() ) <= 1 &&
        std::abs( (vbar(0).left() - b.left()) - (b.right() - vbar(0).right()) ) <= 2 )
      {
      if( hbars() == 1 && 4 * hbar(0).height() <= b.height() )
        {
        if( hbar(0).top() <= b.top() + 1 ) return 'T';
        if( std::abs( hbar(0).vcenter() - b.vcenter() ) <= 1 ) return '+';
        }
      if( hbars() == 2 && 4 * hbar(0).height() <= b.height() &&
          hbar(0).top() <= b.top() + 1 &&
          hbar(1).bottom() >= b.bottom() - 1 &&
          3 * hbar(0).width() > 4 * hbar(1).width() )
        return 'T';
      }
    }

  if( vbars() == 1 && vbar(0).width() >= 2 &&
      2 * vbar(0).width() <= b.width() )
    {
    if( vbar(0).right() <= b.hcenter() )
      {
      if( hbars() == 3 && hbar(0).top() <= b.top() + 1 &&
          hbar(0).width() + 1 >= hbar(1).width() &&
          2 * hbar(1).width() >= 3 * vbar(0).width() &&
          vbar(0).h_overlaps( hbar(1) ) )
        {
        if( Ocrad::similar( hbar(0).width(), hbar(2).width(), 10 ) &&
            hbar(2).width() >= hbar(1).width() &&
            hbar(0).left() <= hbar(1).left() + 1 )
          return 'E';
        if( hbar(0).width() > hbar(2).width() &&
            hbar(1).includes_vcenter( b ) )
          return 'F';
        }
      if( hbars() == 2 )
        {
        if( hbar(0).top() <= b.top() + 1 &&
            hbar(0).width() + 1 >= hbar(1).width() &&
            2 * hbar(1).width() >= 3 * vbar(0).width() &&
            hbar(1).includes_vcenter( b ) )
          return 'F';
        if( hbar(1).bottom() >= b.bottom() - 1 &&
            b.height() > b.width() &&
            hbar(1).width() > hbar(0).width() &&
            std::abs( vbar(0).hcenter() - hbar(0).hcenter() ) <= 1 &&
            rp.iminimum() > rp.pos( 70 ) )
          return 'L';
        }
      if( hbars() == 1 && hbar(0).bottom() >= b.bottom() - 1 &&
          Ocrad::similar( hbar(0).width(), b.width(), 10 ) )
        return 'L';
      }

    if( vbar(0).left() > b.hcenter() &&
        hbars() == 1 && hbar(0).top() <= b.top() + 1 &&
        hbar(0).width() + 1 >= b.width() )
      {
      if( b.width() > b.height() ) return ISO_8859_1::NOT;
      return 0;
      }
    }

  if( vbars() == 1 && vbar(0).width() >= 2 )
    {
    if( 3 * b.height() > 4 * b.width() &&
        std::abs( vbar(0).hcenter() - b.hcenter() ) <= 1 )
      {
      if( rp.istip() && ( lp.istip() || lp.isflats() ) )
        {
        if( lp[lp.pos(40)] > lp[lp.pos(60)] + 1 &&
            rp[rp.pos(60)] > rp[rp.pos(40)] + 1 ) return 'z';
        return 'I';
        }
      if( rp.isflats() && ( lp.istip() || lp.isflats() ) ) return 'l';
      if( rp.istip() && lp.ispit() && Ocrad::similar( lp.iminimum(), rp.pos(50), 10 ) )
        { if( lp.isvpit() ) return '{'; else return '('; }
      if( lp.istip() && rp.ispit() && Ocrad::similar( rp.iminimum(), rp.pos(50), 10 ) )
        { if( rp.isvpit() ) return '}'; else return ')'; }
      if( 2 * b.height() > 3 * b.width() )
        if( rp.isflats() || rp.minima() == 1 )
          if( vbar(0).right() >= b.hpos( 70 ) ||
              b.escape_top( b.vpos( 75 ), vbar(0).right() + 1 ) )
            for( int i = vbar(0).left() - 1; i > b.left(); --i )
              if( b.seek_bottom( b.vpos( 75 ), i ) < b.bottom() &&
                  bp[i-b.left()] <= 1 ) return 'l';
      }
    if( vbar(0).left() <= b.left() + 1 && b.height() > 2 * b.width() &&
        rp.istip() )
      {
      int urow = b.seek_top( b.vcenter(), b.hcenter() );
      int lrow = b.seek_bottom( b.vcenter(), b.hcenter() );
      int ucol = b.seek_right( urow, b.hcenter() );
      int lcol = b.seek_right( lrow, b.hcenter() );
      if( ucol < b.right() && lcol < b.right() )
        return 'C';
      return '[';
      }
    if( vbar(0).right() >= b.right() - 1 )
      {
      if( lp.istip() && b.height() > 2 * b.width() )
        {
        if( 2 * vbar(0).width() <= wp.max() ) return ']';
        if( b.height() >= 3 * b.width() ) return 'l';
        }
      if( 2 * b.height() >= 3 * b.width() &&
          vbar(0).height() >= 3 * vbar(0).width() &&
          lp.isvpit() && lp.minima() == 1 )
        { int i = lp.iminimum();
          if( i > lp.pos(20) && i < lp.pos(40) ) return '1'; }
      }
    }
  if( hbars() == 1 && std::abs( hbar(0).vcenter() - b.vcenter() ) <= 1 &&
      tp.isupit() && bp.isupit() ) return '+';
  return 0;
  }


unsigned char Features::test_c() const throw()
  {
  const Block & b = *_block;

  if( lp.isconvex() )
    {
    if( b.height() > 2 * b.width() ) return '(';
    if( b.seek_bottom( b.vcenter(), b.hcenter() ) < b.bottom() &&
        b.escape_right( b.vcenter(), b.hcenter() ) ) return 'c';
    }
  if( lp.ispit() && bp.ispit() && tp.ispit() &&
      b.escape_right( b.vcenter(), b.hcenter() ) )
    return 'c';
  return 0;
  }


unsigned char Features::test_frst() const throw()
  {
  const Block & b = *_block;

  if( tp.minima( b.height() / 4 ) != 1 || bp.minima( b.height() / 4 ) != 1 )
    return 0;

    {
    int b_hdiff = 0, t_hdiff = 0;
    if( b.bottom_hook( &b_hdiff ) )
      {
      if( -b_hdiff > b.height() / 2 )
        {
        if( b.height() >= 3 * wp.max() &&
            ( hbars() == 0 || hbar(0).bottom() < b.vpos( 20 ) ) ) return 'l';
        if( 3 * wp[1] < b.width() && hbars() >= 1 && hbars() <= 2 &&
            hbar(0).top() >= b.vpos( 20 ) && hbar(0).bottom() < b.vcenter() &&
            Ocrad::similar( hbar(0).width(), wp.max(), 10 ) ) return 't';
        }
      }
    if( b.top_hook( &t_hdiff ) )
      {
      if( 3 * t_hdiff > 2 * b.height() && b.height() > 2 * wp.max() &&
          tp.iminimum() > tp.pos( 50 ) && bp.iminimum() <= bp.pos( 50 ) &&
          ( !b_hdiff || rp.increasing( rp.pos( 50 ) ) ) )
        return 'f';
      if( 2 * b_hdiff > b.height() && 2 * t_hdiff > b.height() ) return 's';
      }
    }

  if( vbars() != 1 || vbar(0).width() < 2 ) return 0;
  if( vbar(0).hcenter() <= b.hcenter() )
    {
    int col = b.right() - rp[b.vcenter()-b.top()] + 2;
    if( col < b.right() )
      {
      int row = b.seek_bottom( b.vcenter(), col );
      if( row >= b.bottom() || b.escape_bottom( row - 1, col ) )
        {
        if( rp.minima() == 3 ) return 'f';
        if( Ocrad::similar( b.height(), b.width(), 40 ) )
          {
          if( tp.minima( b.height() / 8 ) == 2 &&
              bp.minima( b.height() / 8 ) == 2 ) return 'x';
          int row2 = b.vpos(75);
          int col2 = b.seek_right( row2, b.hcenter(), false ) + 1;
          if( b.seek_right( row2, col2 ) >= b.right() )
            {
            if( lp.isconvex() && ( col > b.hcenter() || row < b.bottom() ) )
              return 0;
            if( hbars() == 1 && hbar(0).top() <= b.top() + 1 &&
                4 * hbar(0).height() <= b.height() &&
                4 * lp[lp.pos(50)] >= b.width() ) return 'T';
            return 'r';
            }
          }
        }
      if( Ocrad::similar( b.height(), b.width(), 40 ) )
        {
        if( vscan()[lp.pos(15)] == 3 && vscan()[lp.pos(85)] == 3 &&
            b.seek_right( row - 1, col ) < b.right() && lp.isctip() )
          return 'x';
        if( lp.isvpit() ) return 't';
        }
      }
    if( 3 * b.height() >= 5 * b.width() && vbar(0).left() > b.left() &&
             rp.minima() <= 2 )
      {
      int col = b.right() - rp[b.vcenter() - b.top()] + 1;
      if( !b.escape_bottom( b.vcenter(), col ) )
        { if( tp.ispit() && lp.iminimum() < lp.pos(40) ) return 't'; }
      else if( 2 * wp.max() > b.width() )
        {
        if( rp.iminimum() < rp.pos(20) )
          {
          if( rp.increasing( rp.pos(20) ) || bp.increasing() ||
              tp.minima( 1 ) == 2 ) return 'r';
          else return 'f';
          }
        else return 't';
        }
      }
    if( b.seek_bottom( b.vcenter(), b.hpos( 60 ) + 1 ) >= b.bottom() )
      { if( rp.minima() == 2 ) return 'f'; else return 'r'; }
    }
  return 0;
  }


unsigned char Features::test_G() const throw()
  {
  const Block & b = *_block;

  if( lp.isconvex() || lp.ispit() )
    {
    int col = 0, row = 0;
    for( int i = rp.pos( 40 ); i <= rp.pos( 60 ); ++i )
      if( rp[i] > col ) { col = rp[i]; row = i; }
    if( col == 0 ) return 0;
    row += b.top(); col = b.right() - col + 1;
    if( col <= b.left() || col >= b.hcenter() ) return 0;

    col = ( col + b.hcenter() ) / 2;
    row = b.seek_bottom( row, col );
    if( row < b.bottom() && b.escape_right( row, col ) &&
        !b.escape_bottom( row, b.hcenter() ) )
      {
      int lrow, urow;
      for( lrow = row - 1 ; lrow > b.top(); --lrow )
        if( b.seek_right( lrow, b.hcenter() ) >= b.right() ) break;
      for( urow = lrow - 1 ; urow > b.top(); --urow )
        if( b.seek_right( urow, b.hcenter() ) < b.right() ) break;
      lrow += 2; urow -= 1;
      if( lrow < row && urow > b.top() )
        {
        int uwidth = b.seek_left( urow, b.right() ) - b.seek_right( urow, b.hcenter() );
        int lwidth = b.seek_left( lrow, b.right() ) - b.seek_right( lrow, b.hcenter() );
        if( lrow - 1 <= b.vcenter() || lwidth > uwidth + 2 )
          return 'G';
        }
      }
    }
  return 0;
  }


unsigned char Features::test_HKMNUuvwYy() const throw()
  {
  const Block & b = *_block;

  if( tp.minima( b.height() / 4 ) == 2 && tp.isctip() )
    {
    if( b.height() >= 10 && bp.minima( b.height() / 5 ) == 1 )
      {
      int lg = lp[lp.samples()-2];
      if( lg > 1 && bp.isvpit() && tp.minima( b.height()/2 ) == 2 &&
          lp[lp.pos(75)] <= lg ) return 'v';
      int hdiff;
      if( b.bottom_hook( &hdiff ) )
        {
        if( std::abs( hdiff ) <= b.height() / 8 )
          {
          if( vscan()[b.height()/4] >= 3 ) return 'v';
          if( bp.isconvex() )
            { if( 9 * wp[wp.pos(30)] > 10 * wp[wp.pos(50)] &&
                  9 * wp[wp.pos(50)] > 10 * wp[wp.pos(70)] ) return 'v';
            else return 'u'; }
          }
        if( hdiff > b.height() / 2 ) return 'y';
        }
      int rg = rp[rp.samples()-2];
      if( bp.ispit() && 7 * tp.range() < 4 * b.height() &&
          2 * wp[wp.pos(70)] < b.width() )
        if( ( Ocrad::similar( lg, rg, 20 ) &&
              5 * wp[wp.pos(98)] >= 4 * wp[wp.pos(70)] ) ||
            ( lg > 1 && lg < rg && !rp.increasing() ) ) return 'Y';
      if( 2 * lg < rg && lp.isctip() ) return 'y';
      if( b.escape_top( b.vpos(60), b.hcenter() ) ) return 'u';
      if( lg < rg + 1 && !lp.increasing() && rp.increasing() &&
          ( tp.minima( b.height()/2 ) == 1 || lp.imaximum() > b.height()/2 ) )
        return 'y';
      if( lg > 1 && bp.ispit() && tp.minima( ( 2 * b.height() ) / 5 ) == 2 )
        return 'v';
      if( lg <= 1 && 2 * ( b.width() - rg - lg ) < b.width() &&
          rp.increasing() && tp.minima( b.height()/2 ) == 2 ) return 'v';
      return 0;
      }
    if( bp.minima( b.height() / 4 ) == 2 && bp.isctip() )
      {
      if( vscan()[rp.pos(40)] == 4 ) return 'M';
      int lg = lp[lp.pos(50)];
      int rg = rp[rp.pos(50)];
      if( Ocrad::similar( lg, rg, 80 ) && lg < b.width() / 4 && rg < b.width() / 4 )
        {
        if( lg > 1 && rg > 1 && lp.increasing() && rp.increasing() )
          return 'w';
        if( hbars() == 1 && std::abs( hbar(0).vcenter() - b.vcenter() ) <= 1 &&
            std::abs( hbar(0).hcenter() - b.hcenter() ) <= 1 )
          return 'H';
        if( vscan()[rp.pos(60)] == 4 || vscan()[rp.pos(70)] == 4 ) return 'w';
        if( b.width() > 2 * b.height() ) return '~'; else return 'N';
        }
      if( 3 * lg < 2 * rg && lg < b.width() / 4 && rg > b.width() / 4 &&
          rp.isctip() ) return 'K';
      return 0;
      }
    if( bp.minima() == 3 ) return 'M';
    }
  return 0;
  }


// Looks for the nearest frontier in column hcenter(), then tests if
// gap is open downwards (except for 'x')
unsigned char Features::test_hknwx() const throw()
  {
  const Block & b = *_block;

  if( tp.minima( b.height() / 8 ) == 2 &&
      ( ( lp.istip() && rp.istip() ) ||
        ( lp.isconcave() && rp.isconcave() ) ) ) return 'x';

  int col = 0, row = 0;
  for( int i = bp.pos( 40 ); i <= bp.pos( 60 ); ++i )
    if( bp[i] > row ) { row = bp[i]; col = i; }
  if( row < 2 ) return 0;
  row = b.bottom() - row + 1; col += b.left();
  if( row <= b.top() ) return 0;
  //FIXME
  { int c = col; col = b.seek_right( row, col ); if( col > c ) --col;
  row = b.seek_top( row, col ); }

  if( tp.isconvex() && b.width() >= b.height() && lp.decreasing() ) return '^';
  int urow = b.seek_top( row - 1, col, false );
  if( urow > b.vpos( 20 ) )
    {
    if( tp.minima() >= 2 && tp.minima( b.height() / 5 ) >= 2 ) return 'w';
    if( tp.minima( b.height() / 5 ) == 1 )
      { if( rp.isctip( 66 ) ) return 'k'; else return 'h'; }
    }
  if( Ocrad::similar( b.height(), b.width(), 40 ) && row > b.vcenter() &&
      urow < b.vcenter() && tp.minima( b.height() / 5 ) == 2 &&
      bp.minima( urow + 1 ) == 3 ) return 'w';
  if( urow <= b.vpos( 20 ) && tp.minima( b.height() / 4 ) == 1 &&
      ( b.width() > b.height() || Ocrad::similar( b.height(), b.width(), 40 ) ) )
    return 'n';
  return 0;
  }


bool Features::test_comma() const throw()
  {
  const Block & b = *_block;
  const Blockmap & bm = *_blockmap;

  if( b.block_vector().size() > 0 ) return 0;
  if( 4 * b.height() < 5 * b.width()) return false;
  if( b.height() > 3 * b.width() ) return false;

  if( b.width() >= 3 && b.height() >= 3 )
    {
    int upper_area = 0;
    for( int row = b.top(); row < b.top() + b.width(); ++row )
      for( int col = b.left(); col <= b.right(); ++col )
        if( bm.id( row, col ) == b.id() ) ++upper_area;
    if( upper_area < (b.width() - 2) * (b.width() - 2) ) return false;
    int count1 = 0, count2 = 0;
    for( int col = b.left(); col <= b.right(); ++col )
      { if( bm.id( b.top() + 1, col ) == b.id() ) ++count1;
      if( bm.id( b.bottom() - 1, col ) == b.id() ) ++count2; }
    if( count1 <= count2 ) return false;
    }
  return true;
  }


unsigned char Features::test_easy( int charbox_vcenter ) const throw()
  {
  const Block & b = *_block;
  unsigned char ch = test_solid( charbox_vcenter );
  if( ch ) return ch;

  if( b.top() >= charbox_vcenter && test_comma() ) return ',';
  if( b.bottom() <= charbox_vcenter &&
      b.height() > b.width() && bp.minima() == 1 )
    {
    if( tp.iminimum() < tp.pos(50) && bp.iminimum() > bp.pos(50) )
      return '`'; else return '\'';
    }
  if( 2 * b.height() > 3 * wp.max() && b.top() >= charbox_vcenter &&
      bp.minima() == 1 ) return ',';
  return 0;
  }


// Recognizes single line, non-rectangular characters without holes.
// '/<>\^`
unsigned char Features::test_line( int charbox_vcenter ) const throw()
  {
  const Block & b = *_block;
  int slope1, slope2;

  if( lp.straight( &slope1 ) && rp.straight( &slope2 ) )
    {
    if( slope1 < 0 )
      {
      if( slope2 < 0 ) { if( bp.minima() == 2 ) return '^'; else return 0; }
      if( b.v_includes( charbox_vcenter ) )
        {
        if( 4 * b.area() < b.height() * b.width() ) return '/';
        if( b.height() > b.width() ) return 'l';
        return 0;
        }
      if( b.top() >= charbox_vcenter ) return ',';
      return '\'';
      }
    if( slope1 > 0 && slope2 < 0 )
      {
      if( b.bottom() > charbox_vcenter ) return '\\';
      return '`';
      }
    }
  if( tp.straight( &slope1 ) && bp.straight( &slope2 ) )
    {
    if( slope1 < 0 && slope2 < 0 ) return '<';
    if( slope1 > 0 && slope2 > 0 ) return '>';
    }
  return 0;
  }


unsigned char Features::test_solid( int charbox_vcenter ) const throw()
  {
  const Block & b = *_block;
  if( b.blocks() ) return 0;
  const Blockmap & bm = *_blockmap;
  if( bm.id( b.vcenter(), b.hcenter() ) != b.id() ) return 0;
  int inner_area, inner_space;

  if( b.width() >= 3 && b.height() >= 3 )
    {
    inner_space = ( b.width() - 2 ) * ( b.height() - 2 );
    inner_area = 0;
    for( int row = b.top() + 1; row < b.bottom(); ++row )
      {
      int holes = 0;	// FIXME
      for( int col = b.left() + 1; col < b.right(); ++col )
        { if( bm.id( row, col ) == b.id() ) ++inner_area; else ++holes; }
      if( 5 * holes >= b.width() ) return 0;
      }
    if( inner_area * 100 < inner_space * 75 ) return 0;
    }
  else { inner_space = 0; inner_area = b.area(); }

  if( std::abs( b.height() - wp.max() ) <= 2 ||
      Ocrad::similar( b.height(), wp.max(), 20 ) ) return '.';
  if( inner_area * 100 < inner_space * 85 ) return 0;
  if( b.width() >= 6 * b.height() &&
      b.top() - b.height() > charbox_vcenter ) return '_';
  if( b.width() > b.height() ) return '-';

  if( b.height() > b.width() && tp.minima() == 1 && bp.minima() == 1 )
    {
    if( b.height() > 9 && b.width() <= 8 && b.v_includes( charbox_vcenter ) )
      {
      int l = 0, r = 0;
      for( int row = b.vcenter() - 1; row <= b.vcenter() + 1; ++row )
        { if( bm.id( row, b.left() ) == b.id() ) ++l;
        if( bm.id( row, b.right() ) == b.id() ) ++r; }
      if( l <= 1 && r <= 1 )
        {
        bool lt = false, lb = false, rt = false, rb = false;
        for( int row = b.top(); row <= b.top() + 2; ++row )
          { if( bm.id( row, b.left() ) == b.id() ) lt = true;
          if( bm.id( row, b.right() ) == b.id() ) rt = true; }
        for( int row = b.bottom() - 2; row <= b.bottom(); ++row )
          { if( bm.id( row, b.left() ) == b.id() ) lb = true;
          if( bm.id( row, b.right() ) == b.id() ) rb = true; }
        if( lt && lb && rb ) { if( rt ) return 'I'; else return 'l'; }
        }
      }
    if( b.top() > charbox_vcenter ) return ',';
    if( b.bottom() <= charbox_vcenter ) return '\'';
    return '|';
    }
  return 0;
  }


unsigned char Features::test_misc() const throw()
  {
  const Block & b = *_block;

  if( hbars() == 1 && hbar(0).top() <= b.top() + (b.height() / 10) &&
      5 * hbar(0).width() >= 4 * b.width() &&
      rp.increasing( hbar(0).vcenter() - b.top() ) &&
      rp[hbar(0).bottom()-b.top()+2] - rp[hbar(0).bottom()-b.top()] < b.width() / 4 )
    return '7';

  if( b.height() > b.width() && rp.increasing( 1 ) &&
      b.seek_left( b.vcenter(), b.hcenter() ) <= b.left() )
    return '7';

  if( tp.minima( b.height() / 4 ) == 1 && bp.minima( b.height() / 4 ) == 1 )
    {
    if( b.height() > 2 * wp.max() )
      {
      if( hbars() == 1 && hbar(0).bottom() >= b.bottom() - 1 ) return 'l';
      if( hbars() == 2 && hbar(0).bottom() < b.vpos( 25 ) &&
          hbar(1).bottom() >= b.bottom() - 1 )
        {
        if( 3 * hbar(0).width() <= 2 * hbar(1).width() ) return 'l';
        else return 'I';
        }
      }
    if( hbars() == 1 ||
        ( hbars() == 2 && hbar(1).bottom() >= b.bottom() - 1 &&
        3 * hbar(0).width() > 4 * hbar(1).width() ) )
      if( 3 * hbar(0).height() < b.height() && hbar(0).top() <= b.top() + 1 )
        {
        int i = lp.pos(40);
        if( 3 * wp[i] < b.width() && 5 * lp[i] > b.width() &&
            5 * rp[i] > b.width() ) return 'T';
        }
    if( ( hbars() == 2 || hbars() == 3 ) && hbar(0).top() <= b.top() + 1 &&
        hbar(1).includes_vcenter( b ) &&
        3 * hbar(0).width() > 4 * hbar(1).width() &&
        ( hbars() == 2 ||
          ( hbar(2).bottom() >= b.bottom() - 1 &&
            3 * hbar(0).width() > 4 * hbar(2).width() ) ) ) return 'F';
    }

  if( b.height() > 2 * wp.max() )
    {
    if( rp.istip() && lp.ispit() ) return '{';
    if( lp.istip() && rp.ispit() ) return '}';
    }

  return 0;
  }
